#include <stdio.h>
#include <stdlib.h>

#define C 1
#define CPP 2
#define lenguaje C

#if (lenguaje==C)
	#define bool char
	#define true 1
	#define false 0
#endif

#define soluciones_iniciales true
#define explotar_espejos true
#define solo_simetrico false
#define usar_cotas true


#ifndef N // con el gcc se puede compilar con la opcion -DN=valor
#define N 16
#endif

#define M ((N/2)-1)


//      (mem. reservada) (tiempo de ejecucion)
//
// N=16: 9 Mb Ram         ~segundos
// N=18: 35 Mb Ram        ~30 mintos P2000
// N=20: 180 Mb Ram       ~2,5 dias P2000
// N=22: 759 Mb Ram       ?
//

#define tipo_medicion_tiempo 1
// 1 -> tiempo de CPU (funcion times)
// 2 -> tiempo de reloj (funcion time)
#define pausa_al_acabar 0
// 0 -> no pausa
// 1 -> scanf al final

#if (tipo_medicion_tiempo==1)
#include <sys/times.h>
#elif (tipo_medicion_tiempo==2)
#include <time.h>
#endif

#if (N>=20)
#define smallint short
#else
#define smallint unsigned char
#endif


#define infinito 255 // valido para N<=22 (infinito > coste solucion inicial)

// totalfilas:  suma del total de las posibles combinaciones de terceras filas
//              para cada par de filas (a,b)
// numpares:   numero de combinaciones validas de 2 filas de tamao max_n/2 +1

#if (N==4)
#define totalfilas 148
#define numpares 51
#elif (N==6)
#define totalfilas 705
#define numpares 191
#elif (N==8)
#define totalfilas 3265
#define numpares 699
#elif (N==10)
#define totalfilas 15182
#define numpares 2520
#elif (N==12)
#define totalfilas 71265
#define numpares 9279
#elif (N==14)
#define totalfilas 332487
#define numpares 33926
#elif (N==16)
#define totalfilas 1556631
#define numpares 123890
#elif (N==18)
#define totalfilas 7292741
#define numpares 454162
#elif (N==20)
#define totalfilas 34135770
#define numpares 1661644
#elif (N==22)
#define totalfilas 159943248
#define numpares 6079306
#endif

#ifndef totalfilas
#error N ha de ser par entre 4 y 22
#endif


#include "tipos.h"     // estructuras y tipos
#include "utils.c"   // funciones varias: numero de 0's, invertir, etc
#include "interfic.c" // preguntar datos, mostrar info, mostrar reja
#include "costes.c"  // calculo de costes unarios, binarios y ternarios


costes_binarios costeB1;
// (.muertas.)
// |<---a--->| costeB1[a][b]=coste de la variable a:
// |<---b--->|               *infinito si alguno de los bits (celdas) 0..n/2-1 es inestable
//                           * numero de muertas de los bits 0..n/2-2, si no
// los bits n/2-1, n/2 pertenecen a la columna central de 2xn
costes_unarios costeU;
// (.muertas.)  costeU[a] = infinito si alguna celda 0..n/2-1 de la fila exterior arriba
// |<---a--->|              de a es adyacente a 3 vivas (inestable). 0 si no.
//
// para el pre-proceso de calculo de cotas, estos costes tambien han contado en el numero
// de muertas la celda n/2-1. Despues del pre-proceso, estos costes se vuelven a
// inicializarestan para que no cuenten si esta celda esta muerta o no.
trios t_validos;
//                  bit 0          bit n/2
//                   ^              ^
// trios[c'][ab]:   |<------a------->|
//                  |<------b------->|
//                  |info guardada|c'|
//
// pos[a][b]=ab
// c':  0..3          (2 celdas, centrales, cuyo valor ha sido asignado en el backtracking)
// a,b: 0..2^(n/2+1)-1 (n/2+1 celdas, las 2 ultimas han sido asignadas en el backtracking)
//
// info guardada: coste_ter, num_filas y indice apuntador al vector filas,
//                donde filas[apunt .. apunt+num_filas-1] son todas las combinaciones
//                de terceras filas validas para las dos filas anteriores a y b
//
// modelo equivalente a:
//
//                  bit n/2        bit 0
//                   ^              ^
// trios[ci][ab]:   |<------a------->|
//                  |<------b------->|  // ci: invertida_bits(c')
//                  |ci|info guardada|
//
smallint filas[totalfilas];

posicionador pos; // permite comprimir estructuras con dominio de pares de filas
                  // (elimina combinaciones invalidas)

mejor_extension_cota g_amp;


#include "inis.c"   // funciones de inicializacion excepto cotas
#include "sim.c"    // calculo de la solucion simetrica
#include "cotas.c"  // inicializicacion de cotas


int elims=0;  // variable incrementada en 1 tras cada eliminacion de variable
int elim[N+1]; // lo mismo pero por nivel


#if (!solo_simetrico)


mejor_extension_coste g_izq, g_der;

nuevos_costes_binarios costeBC; // 2 niveles
mejor_extension_fila gf;



// DESCRIPCIN DEL ALGORITMO:
//
// * Bucket-elimination estatico:
// Para cada combinacion posible de celdas de la columna central de 2xN (backtracking)
// realizamos un bucket elimination a la parte izquierda y otro a la parte derecha,
// obteniendo el coste minimo del resto de la reja condicionado a la columna central
// (que tambien tendra un coste)
// Despues, nos quedamos con el mejor valor de coste, y la reja con la que se obtiene ese 
// coste
// * Bucket-elimination dinamico:
// Basicamente lo mismo, con la diferencia que en lugar de eliminar todas las variables
// al final del backtracking, las iremos eliminando a medida que avancemos en el, con
// lo que evitamos repeticiones de eliminaciones de variables en identicas condiciones


// NOTACIN Y CONCEPTOS IMPORTANTES:
//
// para un problema de tamao n:
//  filas: de la 0 a la n-1
//  variables: 1..n (filas) (fila i-1 -> variable i)
//  variables que se eliminan: 3..n
//
//  habiendose eliminado las variables i..n, la variable i-1 tiene aadida
//  a su coste el coste de la mejor extension de las variables i..n
//
// una vez eliminadas las variables 3..n, con un recorrido exhaustivo de las 2 primeras 
// variables obtendremos el coste minimo
//
// para obtener cuales son aquellos valores de las variables, que nos proporcionan un coste
// minimo, lo que hacemos es guardar en cada variable i (i<n), no solamente su coste mas el
// coste de la mejor extension de las variables i+1..n, sino tambien cual es el valor de la
// variable i+1 que nos da ese coste minimo. De esta forma, una vez eliminadas las
// variables 3..n, hallar la solucion optima tiene un coste lineal.
// Ej: var3(var1=54,var2=53) = 3
//     var4(var2=53,var3=3) = 51
//     var5(var3=3,var4=51) = 51
//     ...
// Cada nivel de variable, necesita un espacio de memoria equivalente a D x D , siendo
// D el dominio de las variables
//
// No obstante, en este programa se ha optado por eliminar las variables 4..n y hacer una
// eliminacion especial de la variable 3 en la que ya no se tenga que guardar memoria
// y que directamente nos de el coste minimo y las mejores 3 primeras filas
//
// Entonces:
//            sea i un valor entre 0 y n-4
//            y sean a y b valores del dominio de la variable (0..2^n -1)
//
//            g[i][a][b].valor = el coste de la variable i+3 mas el coste de la mejor 
//                               extension de las variables i+4..n , cuando la variable
//                               i+2 tiene el valor a y la variable i+3 tiene el valor b
//
//            g[i][a][b].arg = el valor de la variable i+4 de la mejor extension de las
//                             variables i+4..n a la variable i+3, cuando la variable
//                             i+2 tiene el valor a y la variable i+3 tiene el valor b
//
//            cada bit del valor de una variable representa una celda, 1=viva, 0=muerta
//
//bit0 bit n/2-2 (total n/2 -1 bits)
// ^   ^
// ----- - -
//|  a  | |
// -----| |  (los valores de g estan condicionados al valor que 
//|  b  | |   ha tomado la fila central en el backtracking)
// -----| |
//|     | |
// ----- - -
//
//            el coste de una variable es un coste ternario, dependiente de la anterior
//            variable y la posterior. Equivale al numero de muertas en los bits 0..n/2
//            de esa variable (no la anterior ni la posterior), o infinito si alguna de
//            las celdas representadas por los bits 0..n/2+1 de esa variable es inestable,
//            o infinito si las 3 celdas representadas por los bits 0 de las 3 variables
//            estan vivas.
//
//            Los costes de la variable 1 y la variable n, puesto que no tienen variables
//            anteriores y posteriores, respectivamente, estan formados
//            -por un coste unario, que indica que la siguiente fila (muertas exteriores a
//             la reja) es toda ella estable (coste unario 0) o no (coste unario infinito),
//            -por un coste binario, que depende de la variable y la adyacente, y equivale
//             a infinito, si alguna de las celdas de la variable es inestable, o su numero
//             de muertas, si no.
//            
// 
//
//
// COTAS (g_amp):
//
//bit0  bit n/2 (total n/2 +1 bits)
// ^     ^
// ------- -
//|  b    |
// -------|  (los valores de g_amp NO estan condicionados al valor que
//|  a    |   ha tomado la fila central en el backtracking)
// -------|
//|       |
// ------- -
//
//           sea i un valor entre 0 y n-4
//           y sean a y b valores del dominio de la variable (0..2^n +1)
//
//           g_amp[i][a][b].valor = el coste de la variable i+2 mas el coste de la mejor 
//                                  extension de las variables 1..i+1 , cuando la variable
//                                  i+2 tiene el valor b y la variable i+3 tiene el valor a
//
//           g_amp[i][a][b].arg = el valor de la variable i+2 de la mejor extension de las
//                                variables 1..i+1 a la variable i+2, cuando la variable
//                                i+2 tiene el valor b y la variable i+3 tiene el valor a
//
//          el coste es el numero de muertas de los bits 0..n/2-1, o infinito si no hay
//          combinacion posible en que todas las celdas representadas por los bits 0..n/2-1
//          sean estables.
//          La ultima celda puede tomar cualquier valor, pero no aadimos su coste, ni
//          podemos mirar su estabilidad (por lo que en las cotas los costes solo difieren
//          en este aspecto).
//
// Entonces:

//  -------------        -------------  
// |    |   |    |      |  *3  |  *5  |
// |----|---|----|      |------|------|
// |    |   |    |      |  *3  |  *5  |
// |----|---|----|      |------|------|
// | ai | a'| ad |      |  *3  |  *5  |      ^
// |----|---|----|  =>  |-------------|      |
// | bi | b'| bd |      | *2 |*1 | *4 |      | asignaciones
// |----|---|----|      |----|---|----|      | de fila central
// |    |   |    |      | *2 |*1 | *4 |      | segun el recorrido
// |----|---|----|      |----|---|----|      | del backtracking
// |    |   |    |      | *2 |*1 | *4 |      |
//  ---- --- ----        ---- --- ---- 
//        ^2 celdas
//
// *1: coste central (todos los costes de las filas centrales asignadas en el backtracking,
//                    a excepcion de la ultima asignada)
// *2: coste de g[j][ai][bi]
// *3: coste de g_amp[j][bi][ai]
//
// el numero de celdas de las cotas de coste minimo va disminuyendo en 1+1 unidad a medida
// que se avanza un nivel en el backtracking
//
// cuando eliminamos la variable j+2 izquierda, no solo se inicializan los valores
// de g[j], sino que se retorna un valor que es una cota minima de coste de *2 + *3
// (es decir, cuando hayamos asignado toda la fila central y eliminado todas las 
// variables izquierda y derecha, el coste minimo que obtengamos sera mayor o igual
// que el proporcionado por esa cota en la parte izquierda + el coste proporcionado
// por la cota en la parte derecha + el coste central, cuando no habiamos eliminado
// todas las variables.
// El valor de la cota de coste minimo es:
// sea m = n/2-1
// cota der = min (x,y pert.a [0.. 2^m -1], X = x+2^m, Y = y+2^m ) de 
//                g[j][x][y]+g_amp[j-1][y][x]
//
// la cota izquierda es la misma, aunque hay que invertir los bits de la fila central
// para que la situacion sea identica)



short eliminar_variable_n (short fcent[N], bool izq, mejor_extension_coste g)
{
// elimina la variable n, y aade a la variable n-1 el coste de la mejor extension
// de la variable n. Este coste depende del valor de la variable n-1 y el valor
// de la variable n-2
// Notacion: variables 1..n -> indices 0..n-1 en fcent
//           variables 3..n-1 -> indices 0..n-4 en g
// Si izq=true, se elimina una variable izquierda, sino se elimina una variable derecha
// Ademas, se devuelve un valor de coste minimo, tal y como se ha explicado
// Solo necesito el coste -> g contiene el coste
//
	short coste;
	short a,b;
	smallint c;
	unsigned char coste_min;
	short max_asig,cmin;
	short col;
	short num_filas,coste_ter;
	short ini_a,ini_b,ini_c;
	short fin_a,fin_b;
	int nf,fin_filas,ab,ba;

	elims++;
	elim[N]++;
	max_asig=(1<<M)-1;

	if (izq)
	{
		// se invierten los bits de fcent para conseguir el modelo invertido equivalente
		ini_a=((fcent[N-3]&2)<<(M-1))+((fcent[N-3]&1)<<(M+1));
		ini_b=((fcent[N-2]&2)<<(M-1))+((fcent[N-2]&1)<<(M+1));
		ini_c=((fcent[N-1]&2)<<(M-1))+((fcent[N-1]&1)<<(M+1));
		col = (fcent[N-1]>>1)+((fcent[N-1]&1)<<1); // la c' en trios[c'][ab]
	}
	else
	{
		// valores de fila central asignados en backtracking, correctamente desplazados
		ini_a=fcent[N-3]<<M;
		ini_b=fcent[N-2]<<M;
		ini_c=fcent[N-1]<<M;
		col = fcent[N-1];
	}
	fin_a=ini_a+max_asig;
	fin_b=ini_b+max_asig;

	cmin=infinito;

	for (a=ini_a;a<=fin_a;a++)
	{
		for (b=ini_b;b<=fin_b;b++)
		{
			ab=pos[a][b]; // elimina combinaciones invalidas (-1) de pares. A posiciones
			              // validas les asigna indices consecutivos
			coste_min=infinito;
			if (ab>=0)
			{
				num_filas=t_validos[col][ab].num_filas;

				if (num_filas>0)
				{
					coste_ter = t_validos[col][ab].coste_ter;
					nf = t_validos[col][ab].apunt;
					fin_filas = nf+num_filas;
					do
					{
						c=filas[nf];
						coste=coste_ter+ costeB1[c+ini_c][b] + costeU[c+ini_c];
						nf++;
						if (coste<coste_min)
						{
							coste_min=(unsigned char)coste;
						}
					}
					while (nf<fin_filas);

					ba=pos[b][a];
					if (coste_min+g_amp[N-4][ba]<cmin)
					{
						cmin=coste_min+g_amp[N-4][ba];
					}
				}
			}
			g[N-4][a-ini_a][b-ini_b]=coste_min;
		}
	}

#if (usar_cotas)
	return cmin;
#else
	return 0;
#endif
}


void eliminar_variable_n_recup_sol (short fcent[N], bool izq, mejor_extension_fila gf,
                                     nuevos_costes_binarios costeBC)
{
// elimina la variable n, y aade a la variable n-1 el coste de la mejor extension
// de la variable n. Este coste depende del valor de la variable n-1 y el valor
// de la variable n-2
// Notacion: variables 1..n -> indices 0..n-1 en fcent
//           variables 3..n-1 -> indices 0..n-4 en g
// Si izq=true, se elimina una variable izquierda, sino se elimina una variable derecha
// Ademas, se devuelve un valor de coste minimo, tal y como se ha explicado
// Necesito los 2 ultimos costes y la fila -> g contiene la fila, costeBC el coste
//
	short coste;
	short a,b;
	smallint c,fila;
	unsigned char coste_min;
	short max_asig;
	short col;
	short num_filas,coste_ter;
	short ini_a,ini_b,ini_c;
	short fin_a,fin_b;
	int nf,fin_filas,ab,ba;

	elims++;
	elim[N]++;
	max_asig=(1<<M)-1;

	if (izq)
	{
		// se invierten los bits de fcent para conseguir el modelo invertido equivalente
		ini_a=((fcent[N-3]&2)<<(M-1))+((fcent[N-3]&1)<<(M+1));
		ini_b=((fcent[N-2]&2)<<(M-1))+((fcent[N-2]&1)<<(M+1));
		ini_c=((fcent[N-1]&2)<<(M-1))+((fcent[N-1]&1)<<(M+1));
		col = (fcent[N-1]>>1)+((fcent[N-1]&1)<<1); // la c' en trios[c'][ab]
	}
	else
	{
		// valores de fila central asignados en backtracking, correctamente desplazados
		ini_a=fcent[N-3]<<M;
		ini_b=fcent[N-2]<<M;
		ini_c=fcent[N-1]<<M;
		col = fcent[N-1];
	}
	fin_a=ini_a+max_asig;
	fin_b=ini_b+max_asig;

	for (a=ini_a;a<=fin_a;a++)
	{
		for (b=ini_b;b<=fin_b;b++)
		{
			ab=pos[a][b]; // elimina combinaciones invalidas (-1) de pares. A posiciones
			              // validas les asigna indices consecutivos
			coste_min=infinito;
			if (ab>=0)
			{
				num_filas=t_validos[col][ab].num_filas;

				if (num_filas>0)
				{
					coste_ter = t_validos[col][ab].coste_ter;
					nf = t_validos[col][ab].apunt;
					fin_filas = nf+num_filas;
					do
					{
						c=filas[nf];
						coste=coste_ter+ costeB1[c+ini_c][b] + costeU[c+ini_c];
						nf++;
						if (coste<coste_min)
						{
							coste_min=(unsigned char)coste;
							fila=c;
						}
					}
					while (nf<fin_filas);

					ba=pos[b][a];
					gf[N-4][a-ini_a][b-ini_b]=fila;
				}
			}
			costeBC[1][a-ini_a][b-ini_b]=coste_min;
		}
	}
}


short eliminar_variable (short i, short fcent[N], bool izq, mejor_extension_coste g)
{
// Pre: eliminadas variables i+1..n
// elimina la variable i, y aade a la variable i-1 el coste de la mejor extension
// de las variables i..n. Este coste depende del valor de la variable i-1 y el valor
// de la variable i-2
// Notacion: variables 1..n -> indices 0..n-1 en fcent
//           variables 3..n-1 -> indices 0..n-4 en g
// Si izq=true, se elimina una variable izquierda, sino se elimina una variable derecha
// Ademas, se devuelve un valor de coste minimo, tal y como se ha explicado
// Solo necesito el coste -> g contiene el coste
//
	short coste;
	short a,b;
	smallint c;
	unsigned char coste_min;
	short max_asig;
	short col;
	short num_filas,coste_ter;
	short cmin,ini_a,ini_b;
	int nf,fin_filas,ab,ba;

	elims++;
	elim[i]++;
	max_asig=(1<<M)-1;

	if (izq)
	{
		// se invierten los bits de fcent para conseguir el modelo invertido equivalente
		ini_a=((fcent[i-3]&2)<<(M-1))+((fcent[i-3]&1)<<(M+1));
		ini_b=((fcent[i-2]&2)<<(M-1))+((fcent[i-2]&1)<<(M+1));
		col = (fcent[i-1]>>1)+((fcent[i-1]&1)<<1); // la c' en trios[c'][ab]
	}
	else
	{
		// valores de fila central asignados en backtracking, correctamente desplazados
		ini_a=fcent[i-3]<<M;
		ini_b=fcent[i-2]<<M;
		col = fcent[i-1];
	}

	cmin=infinito;

	for (a=0;a<=max_asig;a++)
	{
		for (b=0;b<=max_asig;b++)
		{
			ab=pos[a+ini_a][b+ini_b];//elimina combinaciones invalidas (-1) de pares.
			                         //A posiciones validas les asigna indices consecutivos
			coste_min=infinito;
			if (ab>=0)
			{
				num_filas=t_validos[col][ab].num_filas;

				if (num_filas>0)
				{
					coste_ter = t_validos[col][ab].coste_ter;
					nf = t_validos[col][ab].apunt;
					fin_filas = nf+num_filas;
					do
					{
						c=filas[nf];
						coste=coste_ter+g[i-3][b][c];
						nf++;
						if (coste<coste_min)
						{
							coste_min=(unsigned char)coste;
						}
					}
					while (nf<fin_filas);

					ba=pos[b+ini_b][a+ini_a];
					if (coste_min+g_amp[i-4][ba]<cmin)
					{
						cmin=coste_min+g_amp[i-4][ba];
					}
				}
			}
			g[i-4][a][b]=coste_min;
		}
	}

#if (usar_cotas)
	return cmin;
#else
	return 0;
#endif
}


void eliminar_variable_recup_sol (short i, short fcent[N], bool izq, mejor_extension_fila gf,
                                   nuevos_costes_binarios costeBC)
{
// Pre: eliminadas variables i+1..n
// elimina la variable i, y aade a la variable i-1 el coste de la mejor extension
// de las variables i..n. Este coste depende del valor de la variable i-1 y el valor
// de la variable i-2
// Notacion: variables 1..n -> indices 0..n-1 en fcent
//           variables 3..n-1 -> indices 0..n-4 en g
// Si izq=true, se elimina una variable izquierda, sino se elimina una variable derecha
// Ademas, se devuelve un valor de coste minimo, tal y como se ha explicado
// Necesito los 2 ultimos costes y la fila -> g contiene la fila, costeBC el coste
//
	short coste;
	short a,b;
	smallint c,fila;
	unsigned char coste_min;
	short max_asig;
	short col;
	short num_filas,coste_ter;
	short ini_a,ini_b;
	int nf,fin_filas,ab,ba;

	elims++;
	elim[i]++;
	max_asig=(1<<M)-1;

	if (izq)
	{
		// se invierten los bits de fcent para conseguir el modelo invertido equivalente
		ini_a=((fcent[i-3]&2)<<(M-1))+((fcent[i-3]&1)<<(M+1));
		ini_b=((fcent[i-2]&2)<<(M-1))+((fcent[i-2]&1)<<(M+1));
		col = (fcent[i-1]>>1)+((fcent[i-1]&1)<<1); // la c' en trios[c'][ab]
	}
	else
	{
		// valores de fila central asignados en backtracking, correctamente desplazados
		ini_a=fcent[i-3]<<M;
		ini_b=fcent[i-2]<<M;
		col = fcent[i-1];
	}

	for (a=0;a<=max_asig;a++)
	{
		for (b=0;b<=max_asig;b++)
		{
			ab=pos[a+ini_a][b+ini_b];//elimina combinaciones invalidas (-1) de pares.
			                         //A posiciones validas les asigna indices consecutivos
			coste_min=infinito;
			if (ab>=0)
			{
				num_filas=t_validos[col][ab].num_filas;

				if (num_filas>0)
				{
					coste_ter = t_validos[col][ab].coste_ter;
					nf = t_validos[col][ab].apunt;
					fin_filas = nf+num_filas;
					do
					{
						c=filas[nf];
						coste=coste_ter+costeBC[i&1][b][c];
						nf++;
						if (coste<coste_min)
						{
							coste_min=(unsigned char)coste;
							fila=c;
						}
					}
					while (nf<fin_filas);

					ba=pos[b+ini_b][a+ini_a];
					gf[i-4][a][b]=fila;
				}
			}
			costeBC[1-(i&1)][a][b]=coste_min;
		}
	}
}


short calcular_coste_min (short fcent[N], bool izq, mejor_extension_coste g)
{
// Pre: eliminadas variables 4..n
// Eliminacion especial de la variable 3, en la que ya calculo el coste minimo
// de la mejor asignacion de todas las variables 1..n sin guardar ninguna informacion
// adicional en memoria, ademas de obtener las mejores tres primeras filas.
// Notacion: variables 1..n -> indices 0..n-1 en fcent
//           variables 3..n-1 -> indices 0..n-4 en g
// Si izq=true, se elimina variable izquierda, sino se elimina variable derecha
// Solo necesito el coste -> g contiene el coste
//
	short coste;
	unsigned char coste_min;
	short a,b,c;
	short max_asig;
	short col;
	short num_filas;
	short coste_b;
	short ini_a, ini_b;
	int nf,fin_filas,ab;

	elims++;
	elim[3]++;
	max_asig=(1<<M)-1;

	if (izq)
	{
		ini_a=((fcent[0]&2)<<(M-1))+((fcent[0]&1)<<(M+1));
		ini_b=((fcent[1]&2)<<(M-1))+((fcent[1]&1)<<(M+1));
		col = (fcent[2]>>1)+((fcent[2]&1)<<1);
	}
	else
	{
		ini_a=fcent[0]<<M;
		ini_b=fcent[1]<<M;
		col = fcent[2];
	}

	coste_min=infinito;

	for (a=0;a<=max_asig;a++)
	{
		if (costeU[a+ini_a]<infinito)
		{
			for (b=0;b<=max_asig;b++)
			{
				ab=pos[a+ini_a][b+ini_b];
				if (ab>=0)
				{
					coste_b=costeB1[a+ini_a][b+ini_b];
					num_filas=t_validos[col][ab].num_filas;
					if (num_filas>0)
					{
						coste=coste_b+t_validos[col][ab].coste_ter;
						nf = t_validos[col][ab].apunt;
						fin_filas = nf+num_filas;
						do
						{
							c=filas[nf];
							nf++;
							if (coste+g[0][b][c]<coste_min)
							{
								coste_min=coste+g[0][b][c];
							}
						}
						while (nf<fin_filas);
					}
				}
			}
		}
	}

	return coste_min;
}


void encontrar_mejores_3_primeras_filas (short fcent[N], bool izq, 
                                          short *min_a, short *min_b, short *min_c)
{
// Pre: eliminadas variables 4..n
// Eliminacion especial de la variable 3, en la que ya calculo el coste minimo
// de la mejor asignacion de todas las variables 1..n sin guardar ninguna informacion
// adicional en memoria, ademas de obtener las mejores tres primeras filas.
// Si izq=true, se elimina variable izquierda, sino se elimina variable derecha
// Necesito el ultimo coste -> costeBC contiene el coste
//
	short coste;
	unsigned char coste_min;
	short a,b,c;
	short max_asig;
	short col;
	short num_filas;
	short coste_b;
	short ini_a, ini_b;
	int nf,fin_filas,ab;

	elims++;
	elim[3]++;
	max_asig=(1<<M)-1;

	if (izq)
	{
		ini_a=((fcent[0]&2)<<(M-1))+((fcent[0]&1)<<(M+1));
		ini_b=((fcent[1]&2)<<(M-1))+((fcent[1]&1)<<(M+1));
		col = (fcent[2]>>1)+((fcent[2]&1)<<1);
	}
	else
	{
		ini_a=fcent[0]<<M;
		ini_b=fcent[1]<<M;
		col = fcent[2];
	}

	coste_min=infinito;

	for (a=0;a<=max_asig;a++)
	{
		if (costeU[a+ini_a]<infinito)
		{
			for (b=0;b<=max_asig;b++)
			{
				ab=pos[a+ini_a][b+ini_b];
				if (ab>=0)
				{
					coste_b=costeB1[a+ini_a][b+ini_b];
					num_filas=t_validos[col][ab].num_filas;
					if (num_filas>0)
					{
						coste=coste_b+t_validos[col][ab].coste_ter;
						nf = t_validos[col][ab].apunt;
						fin_filas = nf+num_filas;
						do
						{
							c=filas[nf];
							nf++;
							if (coste+costeBC[1][b][c]<coste_min)
							{
								coste_min=coste+costeBC[1][b][c];
								*min_a=a;
								*min_b=b;
								*min_c=c;
							}
						}
						while (nf<fin_filas);
					}
				}
			}
		}
	}
}


void copiar_g (mejor_extension_coste g1, mejor_extension_coste g2, short i)
{
	short max_asig,a,b;

	max_asig=(1<<M)-1;

	for (a=0;a<=max_asig;a++)
		for (b=0;b<=max_asig;b++)
			g2[i][a][b]=g1[i][a][b];
}


short encontrar_mejores_2_filas_centrales (mejor_extension_coste g)
{
	// procedimiento empleado en el caso final recursivo del backtracking
	// cuando hay simetria vertical
//  _____                    _____
// |     |  ^               |     |
// |  a  | _|  equivale a   |  b  | _   cuando la columna central es simetrica vertical
// |  b  |                  |  a  |  |  
// |     |                  |     |  v
//                      
//                                        |x|
//                                        |y|
//                                        |y| : columna simetrica vertical
//                                        |x|
//
	short coste,coste_min;
	short a,b,j;
	short max_asig;

	coste_min=infinito;
	j=(N>>1)-2;
	max_asig=(1<<M)-1;
	
	for (a=0; a<=max_asig; a++)
		for (b=0; b<=max_asig; b++)
		{
			coste = g[j][a][b] + g[j][b][a];
			if (coste<coste_min)
			{
				coste_min=coste;
			}
		}

	return coste_min;
}


bool poda_por_simetria (short j, columnas col)
{
	if (explotar_espejos)
	{
		if (col.izq<col.der)
		{
			return true; // poda por equivalencia con espejo horizontal
		}

		if (j<=(N>>1)-2)
		{
			if (col.abinv<col.arr) return true; //poda por equivalencia con espejo vertical
		}
	}
	return false;
}


void caso_final_dinamico (bool simh, bool simv, bool forzar_sim, short fcent[N], 
                          short coste_cen, short mejor_fcent[N], short *coste_min)
{
	// aquellas acciones a realizar en el caso final recursivo del backtracking, cuando el
	// bucket-elimination es dinamico
	// (basicamente terminar de eliminar las variables y si el coste obtenido es menor
	//  que el optimo, actualizar el optimo y la columna de 2xn que lo causa)
	// El coste_cen es el coste de toda la fila central
	// simh y simv indican si hay simetria vertical u horizantal, respectivamente.
	// coste_min es el mejor coste (la mejor reja) obtenido hasta el momento, coste
	// que se obtiene con la fila central mejor_fcent
	//
	short i,coste_izq,coste_der;

	if (simv)
	{
		coste_izq = encontrar_mejores_2_filas_centrales (g_izq);
	}
	else
	{
		coste_izq = calcular_coste_min(fcent,true,g_izq);
	}

	if (simh)
	{
		coste_der=coste_izq;
	}
	else if (simv)
	{
		coste_der = encontrar_mejores_2_filas_centrales (g_der);
	}
	else
	{
		coste_der = calcular_coste_min(fcent,false,g_der);
	}

	if (coste_izq+coste_cen+coste_der<*coste_min)
	{
		*coste_min=coste_izq+coste_cen+coste_der;
		for (i=0;i<N;i++) mejor_fcent[i]=fcent[i];
		if (!forzar_sim) mantener_info_minimo (*coste_min,false);
	}
}


bool eliminar_variables_y_calcular_costes 
(
	short j, short fcent[N], short coste_cen, short coste_min,
	coste *coste_padre, bool simh, bool forzar_sim, mejor_extension_coste g_izq,
	mejor_extension_coste g_der
)
{
// (en bucket elimination dinamico)                 este comentario no esta bien
// Pre: j>=0
//      inicializadas filas centrales j+1 ..n
//      eliminadas variables j+5 ..n
//
// Elimina variable j y retorna true si el coste minimo que podemos conseguir ya es
// superior al coste minimo que hemos conseguido hasta el momento (por lo tanto se podara)
// coste_cen es el coste de las filas centrales j+2 ..n
//
	short coste_min_acum;
	short coste_min_izq,coste_min_der;

	if (j==N-4)
	{
		coste_min_izq = eliminar_variable_n (fcent,true,g_izq);

		if (forzar_sim)
		{
			coste_min_der=coste_min_izq;
		}
		else if (simh)
		{
			copiar_g (g_izq,g_der,j); // simetria horizontal
			coste_min_der=coste_min_izq;
		}
		else coste_min_der = eliminar_variable_n (fcent,false,g_der);
	}
	else if (j<N-4)
	{
		coste_min_izq = eliminar_variable (j+4,fcent,true,g_izq);

		// La ultima fila central que hemos inicializado no esta aadida en el coste
		// central, sino la mitad izquierda de esa fila (1 celda) en el coste_min_izq,
		// y la mitad derecha de esa celda en el coste_min_der.
		// Por lo que para el padre la fcent[j+2] estaba inicializada, pero su coste
		// estaba incluido en sus coste_min_izq y coste_min_der
		// (mirar explicacion y dibujo lineas 240..290 para aclararse a continuacion,
		//  simular en el dibujo que se avanza un nivel mas en el backtracking)
		//
		if ((fcent[j+2]==1)||(fcent[j+2]==3)) // -> la celda derecha estaba viva
		{
			// el coste_min_der no puede disminuir respecto el del padre,
			// ya que el del padre tenia menos restricciones, y la celda extra que
			// contenia estaba viva tenia un coste 0
			// (coste_min_der>=coste_padre->min_der)
			if (coste_min_izq+coste_cen+coste_padre->min_der >= coste_min) return true;
		}
		else
		{
			// el coste_min_der puede disminuir como maximo en una unidad, ya que habia
			// una celda muerta fija aadida en el coste del padre que ahora ya no se
			// tiene aadida
			// (coste_min_der+1>=coste_padre->min_der)
			if (coste_min_izq+coste_cen-1+coste_padre->min_der >= coste_min) return true;
		}

		if (forzar_sim)
		{
			coste_min_der=coste_min_izq;
		}
		else if (simh)
		{
			copiar_g (g_izq,g_der,j); // simetria horizontal
			coste_min_der=coste_min_izq;
		}
		else coste_min_der = eliminar_variable (j+4,fcent,false,g_der);
	}
	else
	{
		coste_min_izq=0;
		coste_min_der=0;
	}

	coste_min_acum=coste_min_izq+coste_cen+coste_min_der;
	// coste minimo estimado, coste real, coste minimo estimado

	coste_padre->cen=coste_cen;
	coste_padre->min_izq=coste_min_izq;
	coste_padre->min_der=coste_min_der;

	return (coste_min_acum>=coste_min); 
}


void actualizar_columnas (short j, columnas *col, short fcent[N])
{
// actualiza algunos valores (parte izquierda, parte derecha, etc) de la columna central,
// donde cada celda es representada por un bit (1=viva, 0=muerta), en un determinado orden
// Estos valores se utilizan para podar por equivalencia con sus espejos o reflejos.
// Tambien se utilizan para hacer menos trabajo en caso de simetria hor. o vert.
	short desp;

	col->izq = col->izq+(((fcent[j]&2)>>1)<<j);
	col->der = col->der+((fcent[j]&1)<<j);
	if (j>=(N>>1))
	{
		desp=(N-j-1)<<1;
		col->abinv = col->abinv+(fcent[j]<<desp);
	}
	else
	{
		desp=j<<1;
		col->arr = col->arr+(fcent[j]<<desp);
	}
	if (j>=N-7)
	{
		desp=(N-j-1)<<1;
		col->ult2x7 = col->ult2x7+(fcent[j]<<desp);
	}
	else
	{
		col->ult2x7 = col->ult2x7>>2;
		col->ult2x7 = col->ult2x7+(fcent[j]<<12);
	}
}


void backt2col (short j, short fcent[N], short coste_cen, coste coste_padre, columnas col,
                mejor_extension_coste g_izq, mejor_extension_coste g_der, bool forzar_sim,
				short *coste_min, short mejor_fcent[N], bool col_validas[1<<14])
{
	short num;
	bool poda1,poda2,poda3,simh,simv;


	if (j<N-1) actualizar_columnas (j+1,&col,fcent);
	poda1 = poda_por_simetria(j,col); // espejos
	poda2 = !col_validas[col.ult2x7]; //combinaciones con las que no se puede llegar a nada

	if ((!poda1)&&(!poda2))
	{
		simh=(col.izq==col.der);

		if (j>=0)
		{
			if (j<=N-3)
			{
				num=(fcent[j+2]+1)/2;
				coste_cen=coste_cen+2-num;
				num=num+(fcent[j+1]+1)/2;
			}
			else num=0;

			poda3 = eliminar_variables_y_calcular_costes (j,fcent,coste_cen,*coste_min,&coste_padre,simh,forzar_sim,g_izq,g_der);

			if (!poda3)
			{
				// LLAMADAS RECURSIVAS
				// todas las combinaciones de 2x3 celdas en que hay 4 vivas o menos,
				// son validas. Aquellas donde haya 5 o mas, son invalidas

				if (num<3) //num de vivas del anterior 2x2
				{
					fcent[j]=3; // <=2, + 2 celdas -> comb. 2x3 ok
					backt2col (j-1,fcent,coste_cen,coste_padre,col,g_izq,g_der,forzar_sim,coste_min,mejor_fcent,col_validas);
				}
				if ((!forzar_sim)&&(num<4))
				{
					fcent[j]=2; // <=3, + 1 celda -> comb. 2x3 ok
					backt2col (j-1,fcent,coste_cen,coste_padre,col,g_izq,g_der,forzar_sim,coste_min,mejor_fcent,col_validas);

					fcent[j]=1; // <=3, + 1 celda -> comb. 2x3 ok
					backt2col (j-1,fcent,coste_cen,coste_padre,col,g_izq,g_der,forzar_sim,coste_min,mejor_fcent,col_validas);
				}
				{
					fcent[j]=0;
					backt2col (j-1,fcent,coste_cen,coste_padre,col,g_izq,g_der,forzar_sim,coste_min,mejor_fcent,col_validas);
				}
			}
		}
		else // CASOS BASE
		{
			simv=(col.arr==col.abinv);
			coste_cen=coste_cen+4-(fcent[0]+1)/2-(fcent[1]+1)/2;

			caso_final_dinamico (simh,simv,forzar_sim,fcent,coste_cen,mejor_fcent,coste_min);
		}
	}
}


void recuperar_solucion (short fcent[N], int sol[N], mejor_extension_fila gf,
                         nuevos_costes_binarios costeBC)
{
	// tenemos la mejor fila central con la que obtenemos un coste optimo:
	// resolvemos un bucket-elimination para la parte izquierda y otro para la derecha
	// (tiempo=~0) y asignamos la solucion.
	//
	short i,a,b,c;

	eliminar_variable_n_recup_sol (fcent,true,gf,costeBC);

	for (i=N-1;i>=4;i--)
	{
		eliminar_variable_recup_sol (i,fcent,true,gf,costeBC);
	}

	encontrar_mejores_3_primeras_filas (fcent,true,&a,&b,&c);

	sol[0]=a;
	sol[1]=b;
	sol[2]=c;

	a=b;
	b=c;
	for (i=3;i<N;i++)
	{
		c=gf[i-3][a][b];
		sol[i]=c;
		a=b;
		b=c;
	}

	for (i=0;i<N;i++)
	{
		sol[i]=sol[i]+(invertir(fcent[i],2)<<M);
	}

	eliminar_variable_n_recup_sol (fcent,false,gf,costeBC);

	for (i=N-1;i>=4;i--)
	{
		eliminar_variable_recup_sol (i,fcent,false,gf,costeBC);
	}

	encontrar_mejores_3_primeras_filas (fcent,false,&a,&b,&c);

	sol[0]=sol[0]+(invertir(a,M)<<(M+2));
	sol[1]=sol[1]+(invertir(b,M)<<(M+2));
	sol[2]=sol[2]+(invertir(c,M)<<(M+2));

	a=b;
	b=c;
	for (i=3;i<N;i++)
	{
		c=gf[i-3][a][b];
		sol[i]=sol[i]+(invertir(c,M)<<(M+2));
		a=b;
		b=c;
	}
}


#endif



int main()
{
	short i;
	int sol[N];
	short fcent[N];
	short mejor_fcent[N];
	short coste_min,coste_min_ini;
	int num_3as_filas,num_pares;
	bool col_validas[1<<14];
	columnas col;
	coste coste_padre;
	#if (tipo_medicion_tiempo==1)
	struct tms buf;
	#elif (tipo_medicion_tiempo==2)
	time_t t1,t2;
	#endif
	tiempo t;

	presentacion();
	
	#if (tipo_medicion_tiempo==1)
	times (&buf);
	#elif (tipo_medicion_tiempo==2)
	time (&t1);
	#endif
	
	for (i=0;i<=N;i++) elim[i]=0;
	elims=0;

	num_pares=ini_pares_validos(pos); // de tamao M+2

	num_3as_filas=calcular_trios_validos (t_validos);
	
	if ((soluciones_iniciales)||(solo_simetrico))
	{
		ini_costes_para_sim (costeB1, costeU, g_amp);
		coste_min_ini=calcular_sim (g_amp, sol);
		printf ("\nSimetric optimal cost:\n");
		mantener_info_minimo (coste_min_ini,true);
	}
	else coste_min_ini=N*N;

	coste_min=coste_min_ini;

	#if (!solo_simetrico)
	{
		col.arr=0;
		col.abinv=0;
		col.der=0;
		col.izq=0;
		col.ult2x7=0;
		coste_padre.cen=0;

		printf ("Optimal value:\n");

   	ini_costes_para_cotas (costeB1, costeU);
		calcular_cotas (g_amp);
		ini_costes (costeB1, costeU);
		encontrar_cols_validas_de_2x7 (col_validas);

   	backt2col (N-1,fcent,0,coste_padre,col,g_izq,g_der,false,&coste_min,mejor_fcent,col_validas);
		if (coste_min==coste_min_ini) mantener_info_minimo (coste_min,false);

		if (coste_min<coste_min_ini)
		{
			recuperar_solucion (mejor_fcent,sol,gf,costeBC);
		}
	}
	#endif

	#if (tipo_medicion_tiempo==1)
	times (&buf);
	t=tiempo_transcurrido (buf);
	#elif (tipo_medicion_tiempo==2)
	time (&t2);
	t=tiempo_transcurrido (t1,t2);
	#endif

	mostrar_solucion (sol,coste_min,t,elims,elim);

	return 0;
}
